"""
The class that handles parsing input.

:author: Doug Skrypa
"""

from __future__ import annotations

# import logging
from collections import deque
from os import environ
from typing import TYPE_CHECKING, Optional, Union, Any, Deque, List

from .context import ActionPhase, Context, ParseState
from .exceptions import UsageError, ParamUsageError, NoSuchOption, MissingArgument, ParamsMissing
from .exceptions import CommandDefinitionError, Backtrack, UnsupportedAction
from .parse_tree import PosNode
from .parameters.base import BasicActionMixin, Parameter, BasePositional, BaseOption

if TYPE_CHECKING:
    from .command_parameters import CommandParameters
    from .typing import CommandType

__all__ = ['CommandParser']
# log = logging.getLogger(__name__)


class CommandParser:
    """Stateful parser used for a single pass of argument parsing"""

    __slots__ = ('_last', 'arg_deque', 'ctx', 'deferred', 'node', 'params', 'positionals')

    arg_deque: Optional[Deque[str]]
    deferred: Optional[List[str]]
    _last: Optional[Parameter]

    def __init__(self, ctx: Context):
        self._last = None
        self.ctx = ctx
        self.params = ctx.params
        self.positionals = ctx.params.positionals.copy()
        if ctx.config.reject_ambiguous_pos_combos:
            PosNode.build_tree(ctx.command)

    @classmethod
    def parse_args(cls, ctx: Context) -> Optional[CommandType]:
        try:
            parsed = cls.__parse_args(ctx)
        except UsageError:
            ctx.state = ParseState.FAILED
            if not ctx.categorized_action_flags[ActionPhase.PRE_INIT]:
                raise
            return None
        except Exception:
            ctx.state = ParseState.FAILED
            raise
        else:
            if ctx.state == ParseState.INITIAL:
                ctx.state = ParseState.COMPLETE
            return parsed

    @classmethod
    def __parse_args(cls, ctx: Context) -> Optional[CommandType]:
        params = ctx.params
        sub_cmd_param = params.sub_command
        if sub_cmd_param and not sub_cmd_param.choices:
            raise CommandDefinitionError(f'{ctx.command}.{sub_cmd_param.name} = {sub_cmd_param} has no sub Commands')

        cls(ctx)._parse_args(ctx)
        params.validate_groups()

        if sub_cmd_param:
            next_cmd = sub_cmd_param.target()  # type: CommandType
            missing = cls._missing(params, ctx)
            if missing and next_cmd.__class__.parent(next_cmd) is not ctx.command:
                ctx.state = ParseState.FAILED
                if ctx.categorized_action_flags[ActionPhase.PRE_INIT]:
                    return None
                raise ParamsMissing(missing)
            return next_cmd

        missing = cls._missing(params, ctx)
        if missing and not ctx.config.allow_missing and (not params.action or params.action not in missing):
            # Action is excluded because it provides a better error message
            if not ctx.categorized_action_flags[ActionPhase.PRE_INIT]:
                raise ParamsMissing(missing)
        elif ctx.remaining and not ctx.config.ignore_unknown:
            raise NoSuchOption('unrecognized arguments: {}'.format(' '.join(ctx.remaining)))

        return None

    @classmethod
    def _missing(cls, params: CommandParameters, ctx: Context) -> List[Parameter]:
        return [p for p in params.required_check_params() if ctx.num_provided(p) == 0]

    def _parse_args(self, ctx: Context):
        self.arg_deque = arg_deque = self.handle_pass_thru(ctx)
        self.deferred = ctx.remaining = []
        while arg_deque:
            arg = arg_deque.popleft()
            if arg == '--':
                if ctx.params.find_nested_pass_thru():  # pylint: disable=R1723
                    self.deferred.append(arg)
                    self.deferred.extend(arg_deque)
                    break
                else:
                    raise NoSuchOption(f'invalid argument: {arg}')
            elif arg.startswith('---'):
                raise NoSuchOption(f'invalid argument: {arg}')
            elif arg.startswith('--'):
                self.handle_long(arg)
            elif arg.startswith('-') and arg != '-':
                try:
                    self.handle_short(arg)
                except KeyError:
                    self._check_sub_command_options(arg)
                    if self.positionals:
                        try:
                            self.handle_positional(arg)
                        except UsageError:
                            self.deferred.append(arg)
                    else:
                        self.deferred.append(arg)
            else:
                self.handle_positional(arg)

        self._parse_env_vars(ctx)

    def _parse_env_vars(self, ctx: Context):
        # TODO: It would be helpful to store arg provenance for error messages, especially for a conflict between
        #  mutually exclusive params when they were provided via env
        for param in self.params.try_env_params(ctx):
            for env_var in param.env_vars():
                try:
                    value = environ[env_var]
                except KeyError:
                    pass
                else:
                    param.take_action(value)
                    break

    def handle_pass_thru(self, ctx: Context) -> Deque[str]:
        remaining = ctx.remaining
        pass_thru = self.params.pass_thru
        if pass_thru is not None:
            try:
                separator_pos = remaining.index('--')
            except ValueError:
                pass  # If required, it's handled by the normal missing param handler
            else:
                remainder_start = separator_pos + 1
                pass_thru.take_action(remaining[remainder_start:])
                return deque(remaining[:separator_pos])
        return deque(remaining)

    def handle_positional(self, arg: str):
        # log.debug(f'handle_positional({arg=})')
        try:
            param = self.positionals.pop(0)  # type: BasePositional
        except IndexError:
            self.deferred.append(arg)
        else:
            try:
                found = param.take_action(arg)
            except UsageError:
                self.positionals.insert(0, param)
                raise
            try:
                self.consume_values(param, found=found)
            except Backtrack:
                self.positionals.insert(0, param)
            else:
                self._last = param

    def handle_long(self, arg: str):
        # log.debug(f'handle_long({arg=})')
        try:
            opt, param, value = self.params.long_option_to_param_value_pair(arg)
        except KeyError:
            self._check_sub_command_options(arg)
            self.deferred.append(arg)
        else:
            if value is not None or (param.accepts_none and not param.accepts_values):
                param.take_action(value, opt_str=opt)
            elif not self.consume_values(param) and param.accepts_none:
                param.take_action(None, opt_str=opt)
            self._last = param

    def handle_short(self, arg: str):
        # log.debug(f'handle_short({arg=})')
        param_val_combos = self.params.short_option_to_param_value_pairs(arg)
        # log.debug(f'Split {arg=} into {param_val_combos=}')
        if len(param_val_combos) == 1:
            opt, param, value = param_val_combos[0]
            self._handle_short_value(opt, param, value)
        else:
            last_opt, last_param, _last_val = param_val_combos[-1]
            for opt, param, _ in param_val_combos[:-1]:
                param.take_action(None, short_combo=True, opt_str=opt)

            self._handle_short_value(last_opt, last_param, None)

    def _handle_short_value(self, opt: str, param: BaseOption, value: Any):
        # log.debug(f'Handling short {value=} for {param=}')
        if value is not None or (param.accepts_none and not param.accepts_values):
            param.take_action(value, short_combo=True, opt_str=opt)
        elif not self.consume_values(param) and param.accepts_none:
            param.take_action(None, short_combo=True, opt_str=opt)
        self._last = param
        # No need to raise MissingArgument if values were not consumed - consume_values handles checking nargs

    def _check_sub_command_options(self, arg: str):
        # log.debug(f'_check_sub_command_options({arg=})')
        # This check is only needed when subcommand option values may be misinterpreted as positional values
        if not self.positionals:
            return
        param = self.params.find_nested_option_that_accepts_values(arg)
        if param is not None:
            raise ParamUsageError(param, 'subcommand arguments must be provided after the subcommand')

    def _maybe_backtrack(self, param: Parameter, found: int) -> int:
        """
        If we hit the end of the list of provided argument values, unfulfilled Positional parameters remain, and the
        Parameter being processed accepts a variable number of arguments, then check to see if it's possible to
        backtrack to move some of those values to the remaining positionals.

        :param param: The :class:`.Parameter` that was consuming values when the arg_deque became empty
        :param found: The number of values that were consumed by the given Parameter
        :return: The updated found count, if backtracking was possible, otherwise the unmodified found count
        """
        if not self.ctx.config.allow_backtrack or not self.positionals or found < 2:
            return found

        can_pop = param.can_pop_counts()
        to_pop = _to_pop(self.positionals, can_pop, found - 1)
        if to_pop is None:
            return found

        self.arg_deque.extendleft(reversed(param.pop_last(to_pop)))
        return found - to_pop

    def _maybe_backtrack_last(self, param: Union[BasePositional, BasicActionMixin], found: int):
        """
        Similar to :meth:`._maybe_backtrack`, but allows backtracking even after starting to process a Positional.
        """
        if not self.ctx.config.allow_backtrack:
            return

        can_pop = self._last.can_pop_counts()
        to_pop = _to_pop([param, *self.positionals], can_pop, max(can_pop, default=0) + found, found)
        if to_pop is None:
            return

        try:
            reset = param._reset()
        except UnsupportedAction:
            return

        self.arg_deque.extendleft(reversed(reset))
        self.arg_deque.extendleft(reversed(self._last.pop_last(to_pop)))
        raise Backtrack

    def consume_values(self, param: Parameter, found: int = 0) -> int:
        """
        Consume values for the given Parameter.

        :param param: The active :class:`.Parameter` that should receive the discovered values
        :param found: The number of already discovered values for that Parameter (only specified for positional params)
        :return: The total number of values that were found for the given Parameter.
        """
        while True:
            try:
                value = self.arg_deque.popleft()
            except IndexError:
                # log.debug(f'Ran out of values in deque while processing {param=}')
                found = self._maybe_backtrack(param, found)
                return self._finalize_consume(param, None, found)
            else:
                # log.debug(f'Found {value=} in deque - may use it for {param=}')
                if value.startswith('--'):
                    return self._finalize_consume(param, value, found)
                elif value.startswith('-') and value != '-':
                    if self.params.get_option_param_value_pairs(value):
                        # log.debug(f'{value=} will not be used with {param=} - it is also a parameter')
                        return self._finalize_consume(param, value, found)
                    else:
                        try:
                            self._check_sub_command_options(value)
                        except ParamUsageError as e:
                            return self._finalize_consume(param, value, found, e)

                    if not param.would_accept(value):
                        # log.debug(f'{value=} will not be used with {param=} - it would not be accepted')
                        return self._finalize_consume(param, value, found, NoSuchOption(f'invalid argument: {value}'))
                    # log.debug(f'{value=} may be used with {param=} as a value')

                try:
                    found += param.take_action(value)
                except UsageError as e:
                    # log.debug(f'{value=} was rejected by {param=}', exc_info=True)
                    return self._finalize_consume(param, value, found, e)

    def _finalize_consume(
        self, param: Parameter, value: Optional[str], found: int, exc: Optional[Exception] = None
    ) -> int:
        if param.nargs.satisfied(found):
            # Even if an exception was passed to this method, if the found number of values is acceptable, then it
            # doesn't need to be raised.  The value that (would have) caused the exception is added back to the deque.
            if value is not None:
                self.arg_deque.appendleft(value)
            # log.debug(f'consume_values {found=} for {param=}')
            return found
        elif exc:
            raise exc
        elif self._last and isinstance(param, BasePositional) and hasattr(param, '_reset'):
            self._maybe_backtrack_last(param, found)

        n = param.nargs.min
        s = '' if n == 1 else 's'
        raise MissingArgument(param, f'expected {n} value{s}, but only found {found}')


def _to_pop(positionals: List[BasePositional], can_pop: List[int], available: int, req_mod: int = 0) -> Optional[int]:
    if not can_pop:
        return None

    required = sum(p.nargs.min for p in positionals)
    if available < required:
        return None

    required -= req_mod
    nargs_max_vals = [p.nargs.max for p in positionals]
    acceptable = float('inf') if None in nargs_max_vals else sum(nargs_max_vals)
    for n in can_pop:
        if required <= n <= acceptable:
            return n

    return None



"""
Parameter usage / help text formatters

:author: Doug Skrypa
"""
# pylint: disable=W0613

from __future__ import annotations

from typing import TYPE_CHECKING, Union, Type, Callable, Iterator, Iterable, Tuple, Dict

try:
    from functools import cached_property
except ImportError:
    from ..compat import cached_property

from ..config import SubcommandAliasHelpMode
from ..context import ctx
from ..core import get_config
from ..parameters.base import BasePositional, BaseOption
from ..parameters.choice_map import ChoiceMap, Choice
from ..parameters import ParamGroup, PassThru, TriFlag
from .restructured_text import RstTable
from .utils import format_help_entry, _should_add_default

if TYPE_CHECKING:
    from ..typing import Bool, ParamOrGroup, OptStr

BoolFormatterMap = Dict[bool, Callable[[str], str]]


class ParamHelpFormatter:
    _param_cls_fmt_cls_map = {}
    required_formatter_map: BoolFormatterMap = {False: '[{}]'.format}

    def __init_subclass__(cls, param_cls: Type[ParamOrGroup] = None):  # noqa
        if param_cls is not None:
            cls._param_cls_fmt_cls_map[param_cls] = cls

    @classmethod
    def for_param_cls(cls, param_cls: Type[ParamOrGroup]):
        try:
            return cls._param_cls_fmt_cls_map[param_cls]
        except KeyError:
            pass

        for p_cls, f_cls in reversed(tuple(cls._param_cls_fmt_cls_map.items())):  # tuple() only for 3.7 compatibility
            if issubclass(param_cls, p_cls):
                return f_cls

        return ParamHelpFormatter

    def __new__(cls, param: ParamOrGroup):
        fmt_cls = cls.for_param_cls(param.__class__) if cls is ParamHelpFormatter else cls
        return super().__new__(fmt_cls)

    def __init__(self, param: ParamOrGroup):
        self.param = param

    def wrap_usage(self, text: str) -> str:
        try:
            return self.required_formatter_map[self.param.required](text)
        except KeyError:
            return text

    def format_metavar(self) -> str:
        param = self.param
        if param.metavar:
            return param.metavar
        t = param.type
        if t is not None:
            try:
                config = ctx.config
                metavar = t.format_metavar(config.choice_delim, config.sort_choices)
            except Exception:  # noqa  # pylint: disable=W0703
                pass
            else:
                if metavar is not NotImplemented:
                    return metavar

        if ctx.config.use_type_metavar and t is not None:
            try:
                name = t.__name__
            except AttributeError:
                pass
            else:
                if name != '<lambda>':
                    return name.upper()

        return param.name.upper()

    def format_basic_usage(self) -> str:
        """Format the Parameter for use in the ``usage:`` line"""
        return self.wrap_usage(self.format_usage(True))

    def format_usage(self, include_meta: Bool = False, full: Bool = False, delim: str = ', ') -> str:
        """Format the Parameter for use in both the ``usage:`` line and in the list of Parameters"""
        return self.format_metavar()

    def format_description(self, rst: Bool = False) -> str:
        param = self.param
        description = param.help or ''
        if _should_add_default(param.default, description, param.show_default):
            pad = ' ' if description else ''
            quote = '``' if rst else ''
            description += f'{pad}(default: {quote}{param.default!r}{quote})'

        return description

    def format_help(self, prefix: str = '', tw_offset: int = 0) -> str:
        usage = self.format_usage(include_meta=True, full=True)
        description = self.format_description()
        return format_help_entry(usage, description, tw_offset=tw_offset, prefix=prefix)

    def rst_row(self) -> Tuple[str, str]:
        usage = self.format_usage(include_meta=True, full=True)
        return f'``{usage}``', self.format_description(rst=True)


class PositionalHelpFormatter(ParamHelpFormatter, param_cls=BasePositional):
    def format_usage(self, include_meta: Bool = False, full: Bool = False, delim: str = ', ') -> str:
        metavar = self.format_metavar()
        return metavar if not full or self.param.nargs == 1 else f'{metavar} [{metavar} ...]'


class OptionHelpFormatter(ParamHelpFormatter, param_cls=BaseOption):
    def _format_usage_metavar(self) -> str:
        metavar = self.format_metavar()
        if 0 in self.param.nargs:
            metavar = f'[{metavar}]'
        return metavar

    def format_usage_parts(self, delim: str = ', ') -> Union[str, Tuple[str, ...]]:
        param: BaseOption = self.param
        opts = param.option_strs
        if param.nargs == 0:
            return delim.join(opts.option_strs())

        metavar = self._format_usage_metavar()
        option_strs = tuple(opts.option_strs())
        usage_iter = (f'{opt} {metavar}' for opt in option_strs)
        options = len(option_strs)
        if options > 1 and (options * len(metavar) + 2) > max(78, ctx.terminal_width - 2) / 2:
            last = options - 1
            return tuple(f'{us}{delim}' if i < last else us for i, us in enumerate(usage_iter))

        return delim.join(usage_iter)

    def format_usage(self, include_meta: Bool = False, full: Bool = False, delim: str = ', ') -> str:
        if full:
            parts = self.format_usage_parts(delim)
            return parts if isinstance(parts, str) else ''.join(parts)

        param: BaseOption = self.param
        opt = param.option_strs.display_long[0]
        if not include_meta or param.nargs == 0:
            return opt
        return f'{opt} {self._format_usage_metavar()}'

    def format_help(self, prefix: str = '', tw_offset: int = 0) -> str:
        usage = self.format_usage_parts()
        description = self.format_description()
        return format_help_entry(usage, description, tw_offset=tw_offset, prefix=prefix)


class TriFlagHelpFormatter(OptionHelpFormatter, param_cls=TriFlag):
    def format_usage_parts(self, delim: str = ', ') -> Union[str, Tuple[str, ...]]:
        opts = self.param.option_strs
        primary = delim.join(opts.primary_option_strs())
        alts = delim.join(opts.alt_option_strs())
        return primary, alts

    def format_usage(self, include_meta: Bool = False, full: Bool = False, delim: str = ', ') -> str:
        if full:
            return '{} | {}'.format(*self.format_usage_parts(delim))
        else:
            opts = self.param.option_strs
            return f'{opts.display_long_primary[0]} | {opts.display_long_alt[0]}'


class ChoiceMapHelpFormatter(ParamHelpFormatter, param_cls=ChoiceMap):
    @cached_property
    def choice_groups(self) -> Iterable[ChoiceGroup]:
        return ChoiceGroup.group_choices(self.param.choices.values())

    def format_metavar(self) -> str:
        param: ChoiceMap = self.param
        if param.choices:
            config = ctx.config
            choices = (str(c) for c in (c.choice for cg in self.choice_groups for c in cg.choices) if c is not None)
            if config.sort_choices:
                choices = sorted(choices)
            return '{{{}}}'.format(config.choice_delim.join(choices))
        else:
            return param.metavar or param.name.upper()

    def format_usage(self, include_meta: Bool = False, full: Bool = False, delim: str = ', ') -> str:
        return self.format_metavar()

    def format_help(self, prefix: str = '', tw_offset: int = 0) -> str:
        param: ChoiceMap = self.param
        usage = self.format_usage()
        help_entry = format_help_entry(usage, param.description, 2, tw_offset=tw_offset, prefix=prefix)

        choices = self._format_choices(tw_offset, prefix)
        if ctx.config.sort_choices:
            choices = sorted(choices)

        parts = (
            f'{prefix}{param.title or param._default_title}:',
            help_entry,
            *choices,
            prefix.rstrip(),
        )
        return '\n'.join(parts)

    def _format_choices(self, tw_offset: int = 0, prefix: str = '') -> Iterator[str]:
        mode = ctx.config.cmd_alias_mode or SubcommandAliasHelpMode.ALIAS
        for choice_group in self.choice_groups:
            yield from choice_group.format(mode, tw_offset, prefix)

    def rst_table(self) -> RstTable:
        rows = self._format_rst_rows()
        if ctx.config.sort_choices:
            rows = sorted(rows)

        param = self.param
        table = RstTable(param.title or param._default_title, param.description)
        table.add_rows(rows)
        return table

    def _format_rst_rows(self) -> Iterator[Tuple[str, OptStr]]:
        mode = ctx.config.cmd_alias_mode or SubcommandAliasHelpMode.ALIAS
        for choice_group in self.choice_groups:
            for choice, usage, description in choice_group.prepare(mode):
                yield f'``{usage}``', description


class ChoiceGroup:
    """
    A group of :class:`.Choice` objects from a given :class:`~.choice_map.ChoiceMap` that point to the same target.
    The first discovered Choice for a given target is considered the canonical one.  Subsequent Choices for that target
    are considered aliases.

    Used for formatting help text based on the configured :attr:`.CommandConfig.cmd_alias_mode`.
    """

    __slots__ = ('choice_strs', 'choices')

    def __init__(self, choice: Choice):
        choice_str = choice.choice
        self.choices = [choice]
        self.choice_strs = [choice_str] if choice_str else []

    @classmethod
    def group_choices(cls, choices: Iterable[Choice]) -> Iterable[ChoiceGroup]:
        """
        Processes the given Choices to group them by target and configured help text.  If two choices have the same
        target but different help text values, then they are considered different, so they are not grouped together.

        :param choices: The :class:`.Choice` objects that may contain aliases of each other.
        :return: The :class:`.ChoiceGroup` objects containing the grouped Choices.
        """
        target_choice_map = {}
        for n, choice in enumerate(choices):
            key = (choice.target, n if choice.local else None, choice.help)
            try:
                target_choice_map[key].add(choice)
            except KeyError:
                target_choice_map[key] = cls(choice)

        return target_choice_map.values()

    def add(self, choice: Choice):
        self.choices.append(choice)
        choice_str = choice.choice
        if choice_str:
            self.choice_strs.append(choice_str)

    def format(self, default_mode: SubcommandAliasHelpMode, tw_offset: int = 0, prefix: str = '') -> Iterator[str]:
        """
        :param default_mode: The default :class:`.SubcommandAliasHelpMode` to use if no mode was explicitly configured.
        :param tw_offset: Terminal width offset for text width calculations.
        :param prefix: Prefix to add to every line (primarily intended for use with nested groups).
        :return: Generator that yields formatted help text entries (strings) for the Choices in this group.
        """
        for choice, usage, description in self.prepare(default_mode):
            yield format_help_entry(usage, description, lpad=4, tw_offset=tw_offset, prefix=prefix)

    def prepare(self, default_mode: SubcommandAliasHelpMode) -> Iterator[Tuple[Choice, OptStr, OptStr]]:
        """
        Prepares the choice values and descriptions to use for each Choice in this group based on the configured alias
        mode.

        :param default_mode: The default :class:`.SubcommandAliasHelpMode` to use if no mode was explicitly configured.
        :return: Generator that yields 3-tuples containing the :class:`.Choice` object, the choice string value, and
          the help text / description for that choice / alias.
        """
        first = self.choices[0]
        config = get_config(first.target)
        # If it's not a Command, get_config will return None.  If it is a Command, then it will use its config.  If the
        # alias mode is not set on that target Command, but it is set on its parent, then this will use that parent's
        # setting.
        if config:
            mode = config.cmd_alias_mode or default_mode
        else:
            mode = default_mode

        if mode == SubcommandAliasHelpMode.ALIAS:
            yield from self.prepare_aliases()
        elif mode == SubcommandAliasHelpMode.REPEAT:
            yield from self.prepare_repeated()
        elif mode == SubcommandAliasHelpMode.COMBINE:
            yield self.prepare_combined()
        else:  # Treat as a format string
            yield from self.prepare_aliases(mode)

    def prepare_combined(self) -> Tuple[Choice, OptStr, OptStr]:
        """
        Prepare this group's Choices for inclusion in help text / documentation by combining all aliases into a single
        entry.
        """
        first, choice_strs = self.choices[0], self.choice_strs
        try:
            usage, *additional = choice_strs
        except ValueError:  # choice_strs is empty
            return first, first.format_usage(), first.help

        if additional:
            usage = '{{{}}}'.format('|'.join(choice_strs))

        return first, usage, first.help

    def prepare_aliases(self, format_str: str = 'Alias of: {choice}') -> Iterator[Tuple[Choice, OptStr, OptStr]]:
        """
        Prepare this group's Choices for inclusion in help text / documentation using an alternate description for
        aliases.

        Variables supported in the :paramref:`.format_str`:

        - ``{choice}``: The first ("canonical") choice string for this group
        - ``{alias}``: The alias choice string
        - ``{help}``: The original help text for this Choice / group

        To append a suffix to alias descriptions instead of the default prefix, a mode / format string like the
        following could be used::

            cmd_alias_mode='{help} [Alias of: {choice}]'

        :param format_str: The :ref:`format string <python:formatstrings>` to use as the help text / description for
          aliases.
        :return: Generator that yields 3-tuples containing the :class:`.Choice` object, the choice string value, and
          the help text / description for that choice / alias.
        """
        first = self.choices[0]
        try:
            first_str, *choice_strs = self.choice_strs
        except ValueError:  # choice_strs is empty
            yield first, first.format_usage(), first.help
        else:
            help_str = first.help
            yield first, first_str, help_str
            for choice_str in choice_strs:
                yield first, choice_str, format_str.format(choice=first_str, alias=choice_str, help=help_str)

    def prepare_repeated(self) -> Iterator[Tuple[Choice, OptStr, OptStr]]:
        """
        Prepare this group's Choices for inclusion in help text / documentation with no modifications.  Choices that
        are considered aliases are simply repeated as if they were not aliases.
        """
        for choice in self.choices:
            yield choice, choice.format_usage(), choice.help


class PassThruHelpFormatter(ParamHelpFormatter, param_cls=PassThru):
    required_formatter_map = {True: '-- {}'.format, False: '[-- {}]'.format}


class GroupHelpFormatter(ParamHelpFormatter, param_cls=ParamGroup):  # noqa  # pylint: disable=W0223
    required_formatter_map: BoolFormatterMap = {True: '{{{}}}'.format, False: '[{}]'.format}

    def _get_choice_delim(self) -> str:
        param: ParamGroup = self.param
        if param.mutually_dependent:
            return ' + '
        elif param.mutually_exclusive:
            return ' | '
        else:
            return ', '

    def format_usage(self, include_meta: Bool = False, full: Bool = False, delim: str = ', ') -> str:
        choice_delim = self._get_choice_delim()
        members = choice_delim.join(mem.formatter.format_usage(include_meta, full, delim) for mem in self.param.members)
        return self.wrap_usage(choice_delim.join(members))

    def format_description(self, rst: Bool = False) -> str:
        group = self.param
        if not group.description and not group._name:
            if ctx.config.show_group_type and (group.mutually_exclusive or group.mutually_dependent):
                return 'Mutually {} options'.format('exclusive' if group.mutually_exclusive else 'dependent')
            else:
                return 'Optional arguments'
        else:
            description = group.description or f'{group.name} options'
            if ctx.config.show_group_type and (group.mutually_exclusive or group.mutually_dependent):
                description += ' (mutually {})'.format('exclusive' if group.mutually_exclusive else 'dependent')
            return description

    def _get_spacer(self) -> str:
        group = self.param
        if group.mutually_exclusive:
            return '\u00A6 '  # BROKEN BAR
        elif group.mutually_dependent:
            return '\u2551 '  # BOX DRAWINGS DOUBLE VERTICAL
        else:
            return '\u2502 '  # BOX DRAWINGS LIGHT VERTICAL

    def format_help(self, prefix: str = '', tw_offset: int = 0, clean: Bool = True) -> str:
        """
        Prepare the help text for this group.

        :param prefix: Prefix to add to every line (primarily intended for use with nested groups)
        :param tw_offset: Terminal width offset for text width calculations
        :param clean: If this group only contains other groups or Action or SubCommand parameters, then omit the
          description.
        :return: The formatted help text.
        """
        description = self.format_description()
        parts = [f'{prefix}{description}:']

        if ctx.config.show_group_tree:
            spacer = prefix + self._get_spacer()
            tw_offset += 2
        else:
            spacer = prefix

        nested, params = 0, 0
        for member in self.param.members:
            if not member.show_in_help:
                continue

            if isinstance(member, (ChoiceMap, ParamGroup)):
                nested += 1
                parts.append(spacer.rstrip())  # Add space for readability
            else:
                params += 1
            parts.append(member.formatter.format_help(prefix=spacer, tw_offset=tw_offset))

        if clean and nested and not params:
            parts = parts[2:]  # remove description and the first spacer

        if not parts[-1].endswith('\n'):  # ensure a new line separates sections, but avoid extra lines
            parts.append(spacer.rstrip())

        return '\n'.join(parts)

    def rst_table(self) -> RstTable:
        table = RstTable(self.format_description())
        # TODO: non-nested when config.show_group_tree is False; maybe separate options for rst vs help
        for member in self.param.members:
            if member.show_in_help:
                try:
                    sub_table: RstTable = member.formatter.rst_table()  # noqa
                except AttributeError:
                    table.add_row(*member.formatter.rst_row())
                else:
                    sub_table.show_title = False
                    table.add_row(sub_table.title, str(sub_table))

        return table



"""
:author: Doug Skrypa
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Union, Optional, Collection, Iterable, Iterator, MutableMapping, Dict, Set, Tuple

from .exceptions import AmbiguousParseTree
from .utils import _parse_tree_target_repr

if TYPE_CHECKING:
    from .nargs import Nargs
    from .parameters.base import BasePositional
    from .parameters.choice_map import Choice
    from .typing import OptStr, CommandCls

__all__ = ['PosNode']


class AnyWord:
    __slots__ = ('nargs', 'n', 'remaining')

    nargs: Nargs
    n: int
    remaining: Union[int, float]

    def __init__(self, nargs: Nargs, remaining: Union[int, float, None] = None, n: int = 1):
        self.nargs = nargs
        self.n = n
        if remaining is None:
            self.remaining = float('inf') if nargs.max is None else nargs.max - 1  # -1 since one would be consumed
        else:
            self.remaining = remaining

    def __repr__(self) -> str:
        return f'AnyWord({self.nargs!r}, remaining={self.remaining}, n={self.n})'

    def __add__(self, other: int) -> AnyWord:
        remaining = self.remaining - other
        if remaining < 0:
            raise ValueError(f'Unable to add {other} to {self!r} - remaining={remaining} is invalid')
        return AnyWord(self.nargs, remaining, self.n + other)

    def __eq__(self, other: AnyWord) -> bool:
        try:
            return self.nargs == other.nargs and self.remaining == other.remaining and self.n == other.n
        except AttributeError:
            return False

    def __hash__(self) -> int:
        return hash(self.__class__) ^ hash(self.nargs) ^ hash(self.remaining) ^ hash(self.n)


Word = Union[str, AnyWord, None]
Target = Union['BasePositional', 'CommandCls', None]


class PosNode(MutableMapping[Word, 'PosNode']):
    __slots__ = ('links', 'param', 'parent', 'target', 'word', '_any_word', '_any_node')

    links: Dict[Word, PosNode]
    param: Optional[BasePositional]
    parent: Optional[PosNode]
    target: Target
    word: Word

    def __init__(
        self, word: Word, param: Optional[BasePositional], target: Target = None, parent: Optional[PosNode] = None
    ):
        self.links = {}
        self.param = param
        self.parent = parent
        self.target = target
        self.word = word
        try:
            parent[word] = self
        except TypeError:  # parent was None
            pass

    def link_params(self, recursive: bool = False) -> Set[BasePositional]:
        return set(self._link_params(recursive))

    def _link_params(self, recursive: bool = False) -> Iterator[BasePositional]:
        for node in self.values():
            yield node.param
        if recursive:
            for node in self.values():
                try:
                    unbound = node.word.nargs.max is None
                except AttributeError:  # node.word is not an AnyWord
                    yield from node._link_params(True)
                else:
                    yield from node._link_params(not unbound)

    def nargs_min(self) -> int:
        return sum(p.nargs.min for p in self.link_params(True))

    def nargs_max(self) -> Union[int, float]:
        values = [p.nargs.max for p in self.link_params(True)]
        if None in values:
            return float('inf')
        return sum(values)

    # region AnyWord Methods

    @property
    def any_word(self) -> AnyWord:
        try:
            return self._any_word
        except AttributeError:
            try:
                return self._create_child().word
            except (ValueError, TypeError):
                pass
            raise

    @property
    def any_node(self) -> PosNode:
        try:
            return self._any_node
        except AttributeError:
            try:
                return self._create_child()
            except (ValueError, TypeError):
                pass
            raise

    def has_any(self) -> bool:
        try:
            self.any_word  # noqa
        except AttributeError:
            return False

        return True

    def _create_child(self) -> PosNode:
        # Will raise ValueError if self.word has remaining < 1
        word = self.word + 1
        return PosNode(word, self.param, self.target, self)

    # endregion

    # region Introspection

    @property
    def raw_path(self) -> Tuple[Word, ...]:
        word = self.word
        if not word:
            return ()
        return (*self.parent.raw_path, word)  # noqa

    def path_repr(self) -> str:
        return '({})'.format(', '.join(str(n) if isinstance(n, AnyWord) else repr(n) for n in self.raw_path))

    # endregion

    # region Dunder Methods

    def __repr__(self) -> str:
        if self.param == self.target:
            pt_str = f'param=target={_parse_tree_target_repr(self.target)}'
        else:
            pt_str = f'param={self.param}, target={_parse_tree_target_repr(self.target)}'
        return f'<PosNode[{self.path_repr()}: {self.word!r}, links: {len(self)}, {pt_str}]>'

    def __hash__(self) -> int:
        return hash(self.__class__) ^ hash(self.parent) ^ hash(self.word) ^ hash(self.param) ^ hash(self.target)

    def __eq__(self, other: PosNode) -> bool:
        return (
            self.parent == other.parent
            and self.param == other.param
            and self.word == other.word
            and self.target == other.target
            and self.links == other.links
        )

    def __bool__(self) -> bool:
        if self.links:
            return True
        return self.has_any()

    def __len__(self) -> int:
        return len(self.links) + self.has_any()

    def __contains__(self, word: Word) -> bool:
        try:
            self.links[word]
        except KeyError:
            pass
        else:
            return True
        try:
            return self.any_word == word
        except AttributeError:
            return False

    def __setitem__(self, word: Word, node: PosNode):
        if isinstance(word, AnyWord):
            try:
                self._any_word
            except AttributeError:
                try:
                    next_word = self.word + 1
                except (TypeError, ValueError):
                    pass
                else:
                    if word != next_word or node.parent is not self or node.param is not self.param:
                        raise KeyError(f'Choice conflict: {word!r}=>{node} cannot replace {next_word!r}')

                self._any_word = word
                self._any_node = node
                return

            raise KeyError(f'Choice conflict: {word!r}=>{node} cannot replace {self._any_word!r}=>{self._any_node}')
        else:
            self.links[word] = node

    def __getitem__(self, word: Word) -> PosNode:
        try:
            return self.links[word]
        except KeyError:
            pass
        try:
            return self.any_node
        except AttributeError:
            pass
        raise KeyError(word)

    def __delitem__(self, word: Word):
        try:
            del self.links[word]
        except KeyError:
            pass
        else:
            return
        try:
            any_word = self._any_word
        except AttributeError:
            raise KeyError(word) from None

        if any_word == word:
            del self._any_word
            del self._any_node
        else:
            raise KeyError(word)

    def __iter__(self) -> Iterator[Word]:
        yield from self.links
        try:
            yield self.any_word
        except AttributeError:
            pass

    # endregion

    # region Build Tree

    @classmethod
    def build_tree(cls, command: CommandCls) -> PosNode:
        root = cls(None, None, target=command)
        process_params(command, [root], command.__class__.params(command).positionals)
        return root

    def update_node(self, word: Word, param: BasePositional, target: Target) -> PosNode:
        try:
            *parts, last = word.split()
        except AttributeError:  # The choice is None or Any
            if word:
                return self._update_any(word, param, target)
            else:
                return self._set_target(target)
        else:
            node = self
            for part in parts:
                node = node._update(part, param, None)
            return node._update(last, param, target)

    def _set_target(self, target: Target) -> PosNode:
        if not target:
            return self
        elif self.target and self.word:
            raise AmbiguousParseTree(self, target)
        else:
            self.target = target
            return self

    def _update(self, word: str, param: BasePositional, target: Target) -> PosNode:
        links = self.links
        try:
            node = links[word]
        except KeyError:
            if self.has_any():
                raise AmbiguousParseTree(self, target, word) from None
            return PosNode(word, param, target, self)
        else:
            return node._set_target(target)

    def _update_any(self, word: AnyWord, param: BasePositional, target: Target) -> PosNode:
        try:
            self[word]
        except KeyError:
            pass
        else:
            raise AmbiguousParseTree(self, target, word)

        node = PosNode(word, param, target, self)
        if word.nargs.max is not None:
            while True:
                try:
                    node = node._create_child()
                except ValueError:
                    break

        return node

    # endregion

    def print_tree(self, indent: int = 0, recursive: bool = True):
        prefix = ' ' * indent
        print(f'{prefix}- <PosNode[{self.word!r}, links: {len(self)}, target={self.target!r}]>')
        if not recursive:
            return
        indent += 2
        for node in self.values():
            try:
                unbound = node.word.nargs.max is None
            except AttributeError:  # node.word is not an AnyWord
                node.print_tree(indent, True)
            else:
                node.print_tree(indent, not unbound)


def process_params(command: CommandCls, nodes: Iterable[PosNode], params: Iterable[BasePositional]) -> Set[PosNode]:
    for param in params:
        nodes = process_param(command, nodes, param)

    return nodes


def process_param(command: CommandCls, nodes: Iterable[PosNode], param: BasePositional) -> Set[PosNode]:
    # At each step, the number of branches grows
    try:
        choices: Dict[OptStr, Choice] = param.choices  # noqa
    except AttributeError:  # It was not a ChoiceMap param
        pass
    else:
        get_params = command.__class__.params

        new_nodes = set()
        for choice in choices.values():
            target = choice.target
            try:
                params = get_params(target)
            except TypeError:
                new_nodes.update(node.update_node(choice.choice, param, target) for node in nodes)
            else:
                choice_nodes = {node.update_node(choice.choice, param, target) for node in nodes}
                new_nodes.update(process_params(target, choice_nodes, params.positionals))

        return new_nodes

    try:
        choices: Collection[str] = param.type.choices  # noqa
    except AttributeError:  # It was not a _ChoicesBase input type
        pass
    else:
        return {node.update_node(choice, param, param) for choice in choices for node in nodes}

    # At this point, the param will take any word
    word = AnyWord(param.nargs)
    return {node.update_node(word, param, param) for node in nodes}
